package com.pact;

import java.util.*;

public class Node {

    public static class MatchedNodePair {
        Node node1;
        Node node2;

        MatchedNodePair(Node n1, Node n2) {
            this.node1 = n1;
            this.node2 = n2;
        }

        /**
         * Gets size of shared node (just length of venn)
         * @return size of node
         */
        int nodesSize()
        {
            return this.node1.getVenn().length();
        }

        public String toString() {
            return "<" + this.node1 + ", " + this.node2 + ">";
        }
    }

    private Node parent;
    private String venn;
    private List<Node> children;
    private Node linkNode = null;
    private List<Node> linkNodes = new ArrayList<>();

    private int hash = 0;

    public Node(Node p, String venn) {
        this.venn = venn;
        this.parent = p;

        this.children = new ArrayList<>();
        if (venn.length() > 1)
            setChildren(venn);
        updateHash();
    }

    private Node(Node p, Collection<Node> children) {
        this.parent = p;
        this.children = new ArrayList<>();
        this.children.addAll(children);
        for(Node child : children) {
            Node child_ = child.copy();
            child_.parent = this;
            this.children.add(child_);
        }

        updateVenn();
        updateHash();
    }

    private Node(Node p, char s) {
        this(p, ""+s);
    }

    private void updateHash(){
        hash = 0;
        if (children.isEmpty()) {
            hash = (venn.hashCode()*1073676287) ^ (venn.hashCode()*97);
        } else {
            for (Node child : children) {
                hash = hash ^ child.hash;
            }
            hash *= 33;
        }
    }

    private void updateVenn() {
        StringBuilder sb = new StringBuilder();
        for(Node child : children){
            sb.append(child.venn);
        }
        venn = "(" + sb.toString() + ")";
    }

    private boolean isRoot() {
        return parent == null;
    }

    private boolean isLeaf() {
        return children.isEmpty();
    }

    /**
     * Traverses the tree with root at given node and returns all leaf nodes
     * @return Leaf nodes under this root
     */
    private List<Node> getLeaves() {
        List<Node> result = new ArrayList<>();
        // Use stack to avoid recursion here
        Stack<Node> searchNodes = new Stack<>();
        searchNodes.push(this);
        while (!searchNodes.isEmpty()) {
            Node curr = searchNodes.pop();
            if (curr.isLeaf()) {
                result.add(curr);
            } else {
                searchNodes.addAll(curr.children);
            }
        }
        return result;
    }

    /**
     * Traverses the tree with root at given node and returns the number of leaves found.
     * @return Number of leaf nodes under this root
     */
    private int getNLeaves() {
        int result = 0;
        // Use stack to avoid recursion here
        Stack<Node> searchNodes = new Stack<>();
        searchNodes.push(this);
        while (!searchNodes.isEmpty()) {
            Node curr = searchNodes.pop();
            if (curr.isLeaf()) {
                result++;
            } else {
                searchNodes.addAll(curr.children);
            }
        }
        return result;
    }

    private int nSiblings() {
        if (isRoot()) return 0;
        else return parent.children.size() - 1;
    }

    private void setChildren(String s) {
        try {
            String vennChildren = s.substring(1, s.length() - 1);
            int count = 0;
            while (count < vennChildren.length())
                if (vennChildren.charAt(count) == '(') {
                    int start = count;
                    count++;
                    int end;
                    int brackets = 1;
                    while (true)
                    {
                        if (vennChildren.charAt(count) == '(') {
                            brackets++;
                        }
                        if (vennChildren.charAt(count) == ')') {
                            brackets--;
                            if (brackets == 0) {
                                count++;
                                break;
                            }
                        }
                        count++;
                    }
                    end = count;
                    Node child = new Node(this, vennChildren.substring(start, end));

                    this.children.add(child);
                } else {
                    Node child = new Node(this, vennChildren.charAt(count));

                    this.children.add(child);
                    count++;
                }
        }
        catch (StringIndexOutOfBoundsException se) {
            throw new BracketException();
        }
    }

    String getVenn()
    {
        return this.venn;
    }

    public String toString()
    {
        return this.venn;
    }

    private List<Node> getSiblings() {
        if (this.parent != null) {
            List<Node> sibs = new ArrayList<>(parent.children);
            sibs.remove(this);
            return sibs;
        }
        return new ArrayList<>();
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) return true;  // Literally the same object
        if (!(obj instanceof Node)) return false; // Wrong type

        Node node = (Node) obj;
        if (hash != node.hash) return false; // equal object have equal hash codes

        // Case for leaf nodes
        if (this.isLeaf() && node.isLeaf()) {
            return this.venn.equals(node.venn);
        }

        if (children.size() != node.children.size()){
            return false;
        }

        // If above checks fall through, we need to actually look at the structure
        // Of the Node.


        // Check to make sure all of my children exist in node's children
        for (Node theirChild : node.children) {
            boolean foundMatch = false;
            for (Node myChild : this.children) {
                if (myChild.equals(theirChild)) {
                    foundMatch = true;
                    break;
                }
            }
            if(!foundMatch) return false;
        }
        return true;
    }

    @Override
    public int hashCode(){
        return hash;
    }

    private Node copy(){
        return new Node(parent, children);
    }

    private Node getRoot() {
        if (isRoot()) return this;
        else return this.parent.getRoot();
    }

    /**
     * Search this tree for the subtree with given root
     * @param targetNode the root of the subtree to search for
     */
    private List<MatchedNodePair> search(Node targetNode) {
        List<MatchedNodePair> result = new ArrayList<>();
        // Use stack to avoid recursion here
        Stack<Node> searchNodes = new Stack<>();
        searchNodes.push(this);
        while (!searchNodes.isEmpty()) {
            Node curr = searchNodes.pop();
            if (curr.equals(targetNode)) {
                curr.linkNodes.add(targetNode);
                targetNode.linkNodes.add(curr);
                result.add(new MatchedNodePair(curr, targetNode));
            } else {
                searchNodes.addAll(curr.children);
            }
        }
        return result;
    }

    /**
     * Searches in this tree for the presence of the Node node.
     * @param node root of subtree to search for
     * @return whether node was found in this tree
     */
    private boolean isIn(Node node) {
        if (this == node) {
            return true;
        }
        for (Node child : this.children) {
            if (child.isIn(node)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Looks through this Node's siblings and returns the first one that has linkNodes. If no sibling
     * has linkNodes, returns null.
     * @return the link, null if it doesn't exist
     */
    private Node getLink() {
        for (Node curr : getSiblings()) {
            if (!curr.linkNodes.isEmpty()) {
                return curr;
            }
        }
        return null;
    }

    private boolean hasLink() {
        return getLink() != null;
    }

    private boolean unshared() {
        if (this.linkNodes.isEmpty()) {
            if(this.children.isEmpty()) {
                return true;
            } else {
                for (Node node : this.children) {
                    if (!node.unshared()) {
                        return false;
                    }
                }
            }
            return true;
        } else {
            return false;
        }
    }

    private void resetSearch() {
        this.linkNodes.clear();
        this.linkNode = null;

        for (Node child : this.children) {
            child.resetSearch();
        }
    }

    /**
     * Searches in h for all NodePairs that contain this object as either node1 or node2
     * @param matchedNodePairs collection to search for duplicates in
     * @return a copy of the list with only duplicates remaining
     */
    private List<Node> getDuplicates(List<MatchedNodePair> matchedNodePairs) {
        List<Node> result = new ArrayList<>();
        for (MatchedNodePair nodePair : matchedNodePairs) {
            if (nodePair.node1 == this)
                result.add(nodePair.node2);
            else if (nodePair.node2 == this) {
                result.add(nodePair.node1);
            }
        }
        return result;
    }

    /**
     * Traverses tree and collects all unshared nodes.
     * @return unshared nodes
     */
    private List<Node> collectUnshared() {
        Stack<Node> stack = new Stack<>();
        List<Node> unshared = new ArrayList<>();
        stack.push(this);
        while (!stack.isEmpty()) {
            Node n = stack.pop();
            if (n.unshared()) {
                unshared.add(n);
            }
            stack.addAll(n.children);
        }
        return unshared;
    }

    /**
     * Produces a reduced version of unreduced where any duplicate children have been removed. Note that this doesn't
     * look for duplicate sub-trees at lower levels. (ie duplicate grandchildren)
     * @param unreduced The unreduced tree
     * @return Node with duplicate children removed
     */
    private static Node reduce(Node unreduced) {
        if(unreduced.isLeaf()) {
            // Cannot reduce a leaf, so return unmodified
            return unreduced;
        }

        HashSet<Node> uniqueChildren = new HashSet<>();
        for (Node child : unreduced.children) {
            if (!uniqueChildren.contains(child)) {
                uniqueChildren.add(child);
            }
        }
        if (uniqueChildren.size() == unreduced.children.size()) {
            // There are no duplicate children so return unmodified
            return unreduced;
        }
        Iterator<Node> uIt = uniqueChildren.iterator();

        StringBuilder result = new StringBuilder();
        result.append(uIt.next().getVenn());
        if (uIt.hasNext()) {
            while (uIt.hasNext()) {
                result.append(uIt.next().getVenn());
            }
            result.insert(0, "(");
            result.append(")");
        }
        return new Node(null, result.toString());
    }

    public static List<MatchedNodePair> deepSearch(Node nodeA, Node nodeB) {
        nodeA.resetSearch();
        nodeB.resetSearch();
        List<MatchedNodePair> result = new ArrayList<>();

        Stack<Node> searchNodes = new Stack<>();
        searchNodes.push(nodeB);
        while (!searchNodes.isEmpty()) {
            Node curr = searchNodes.pop();

            // res1: Occurrences of node in this tree
            List<MatchedNodePair> res1 = nodeA.search(curr);

            // res2: Occurrences of this tree in node
            List<MatchedNodePair> res2 = curr.search(nodeA);

            assert(res1.size() == 0 || res2.size() == 0 || res1.equals(res2));

            if(!res1.isEmpty()){
                result.addAll(res1);
            } else if(!res2.isEmpty()){
                result.addAll(res2);
            }
            searchNodes.addAll(curr.children);
        }
        return result;
    }

    private static List<Node> combine(Node node1, Node node2) {
        log("Combining: " + node1 + ", " + node2);
        List<Node> result = _combine(node1, node2);
        result = removeDuplicates(result);
        log("Combining Result: " + result);
        return result;
    }

    private static List<Node> _combine(Node node1, Node node2) {
        List<Node> result = new ArrayList<>();

        // Step1. Remove any duplicate children of node1 and node2.
        node1 = reduce(node1);
        node2 = reduce(node2);

        // Case of the two reduced nodes being identical, no extra work required, just return one of them.
        if (node1.equals(node2)) {
            result.add(node2);
            return result;
        }



        List<MatchedNodePair> sResult = deepSearch(node2, node1);

        List<Node> unsharedNodes = new ArrayList<>();
        unsharedNodes.addAll(node2.collectUnshared());
        unsharedNodes.addAll(node1.collectUnshared());

        if ((unsharedNodes.isEmpty()) && (node1.getNLeaves() == node2.getNLeaves())) {
            if (node1.children.size() < node2.children.size()) {
                Node toAdd = new Node(null, node1.getVenn());
                result.add(toAdd);
                return result;
            } else if (node2.children.size() < node1.children.size()) {
                Node toAdd = new Node(null, node2.getVenn());
                result.add(toAdd);
                return result;
            }

        }

        List<MatchedNodePair> largest = getLargestNodePairs(sResult);
        if (!largest.isEmpty()) {
            Iterator itLargest = largest.iterator();

            MatchedNodePair flargest = (MatchedNodePair)itLargest.next();

            List<Node> lg1sibs = flargest.node1.getSiblings();
            List<Node> lg2sibs = flargest.node2.getSiblings();

            List<Node> newSibs = new ArrayList<>();
            List<Node> otherSibs1 = new ArrayList<>();
            List<Node> otherSibs2 = new ArrayList<>();

            sortLargestSibs(lg1sibs, flargest.node2, newSibs, otherSibs1);
            sortLargestSibs(lg2sibs, flargest.node1, newSibs, otherSibs2);


            if ((otherSibs1.isEmpty()) && (otherSibs2.isEmpty())) {
                unsharedNodes.addAll(newSibs);
            }

        }

        List<MatchedNodePair> newLinkedNodePair = getLinkedNodePair(unsharedNodes);

        List<List<MatchedNodePair>> polys = reviewPolytomies(newLinkedNodePair);
        if (!polys.isEmpty()) {
            Iterator<List<MatchedNodePair>> polyIt = polys.iterator();
            if (polyIt.hasNext()) {
                List<MatchedNodePair> polyresult = addPolys(polyIt.next());
                Iterator prIt = polyresult.iterator();
                if (prIt.hasNext()) {
                    MatchedNodePair curr = (MatchedNodePair)prIt.next();
                    if ((curr.node1.parent != null) && (curr.node1.linkNode.parent != null)) {
                        String polyString1 = buildUp(curr.node1.parent, curr.node2.getVenn());
                        String polyString2 = buildUp(curr.node1.linkNode.parent, curr.node2.getVenn());
                        Node polyOne = new Node(null, polyString1);
                        Node polyTwo = new Node(null, polyString2);
                        result.addAll(combine(polyOne, polyTwo));
                    } else if (!curr.node1.isRoot()) {
                        String polyString1 = buildUp(curr.node1.parent, curr.node2.getVenn());
                        Node polyOne = new Node(null, polyString1);
                        result.add(polyOne);
                    } else if (!curr.node1.linkNode.isRoot()) {
                        String polyString2 = buildUp(curr.node1.linkNode.parent, curr.node2.getVenn());
                        Node polyTwo = new Node(null, polyString2);
                        result.add(polyTwo);
                    } else {
                        throw new RuntimeException("");
                    }
                }
            }
            return result;
        }

        if (!newLinkedNodePair.isEmpty()) {
            List<MatchedNodePair> combined = combineNew(newLinkedNodePair.get(0));
            for (MatchedNodePair curr : combined) {
                result.addAll(combine(curr.node1, curr.node2));
            }
            return result;
        }

        if (largest.get(0).nodesSize() == 1) { // Shares only leaves
            List<Node> unique = getUniqueNodes(sResult);

            for (Node curr : unique) {
                List<Node> dupres = curr.getDuplicates(sResult);
                if (dupres.size() > 1) {
                    for (MatchedNodePair aNewDupNodePair : getLinkedNodePair(dupres)) {
                        for (MatchedNodePair curr2 : combineNew(aNewDupNodePair)) {
                            List<Node> newResults = combine(curr2.node1, curr2.node2);
                            result.addAll(newResults);
                        }
                    }
                }
                else {
                    result.add(combineAtRoot(node2, node1));
                }
            }
            return result;
        }
        else {  // shares sub-trees
            for (MatchedNodePair largest_ : largest) {
                result.addAll(subTrees(largest_));
            }
            return result;
        }
    }

    public static List<Node> reducedCombine(Node node1, Node node2) {
        log("Attempting to combine " + node1 + " and " + node2);
        List<Node> result = new ArrayList<>();

        int nSharedLeaves = sharedLeaves(node1, node2).size();
        if (nSharedLeaves <= 1) {
            log(String.format("Found %d shared leaves, combining at root", nSharedLeaves));
            result.add(combineAtRoot(node1, node2));
        } else {
            log(String.format("Found %d shared leaves, running full combine", nSharedLeaves));
            List<Node> compCombine = combine(node1, node2);
            log(String.format("combine yielded %d unique trees, reducing to smallest.", compCombine.size()));

            int smallest = Integer.MAX_VALUE;
            for (Node curr : compCombine) {
                int currAreas = curr.getLeaves().size();
                if (currAreas < smallest) {
                    smallest = currAreas;
                    result.clear();
                    result.add(curr);
                } else if (currAreas == smallest) {
                    result.add(curr);
                }
            }
            log(String.format("After reduction, there are %d unique trees with size %d.", result.size(), result.get(0).getNLeaves()));
        }
        log("Result of Combination is " + result);
        return result;
    }

    /**
     * Removes any duplicate nodes. Guarantees that
     * forall A, B in result, A.equals(B) == false
     * @param all List of nodes from which to remove duplicates
     * @return all with duplicates removed
     */
    private static List<Node> removeDuplicates(List<Node> all) {
        // TODO: Speed up w/ sets as soon as hash problems are sorted
        List<Node> result = new ArrayList<>();
        for (Node curr : all) {
            boolean foundDup = false;
            for (Node resNode : result) {
                if (resNode.equals(curr)) {
                    foundDup = true;
                    break;
                }
            }
            if (!foundDup) {
                result.add(curr);
            }
        }
        return result;
    }

    /**
     * Returns all leaf nodes that are shared by the two input trees.
     * @param node1 Tree to search for shared leaves
     * @param node2 Tree to search for shared leaves
     * @return The list of shared nodes (Actual nodes belong to the tree node1)
     */
    private static List<Node> sharedLeaves(Node node1, Node node2) {
        Set<Node> shared = new HashSet<>(node1.getLeaves());
        shared.retainAll(node2.getLeaves());
        return new ArrayList<>(shared);
    }

    private static void sortLargestSibs(List<Node> lgsibs, Node lg, List<Node> newSibs, List<Node> os) {
        for (Node curr : lgsibs) {
            Iterator linkIt = curr.linkNodes.iterator();
            boolean linkFlargest = false;
            if (linkIt.hasNext()) {
                linkFlargest = true;
            }
            while (linkIt.hasNext()) {
                Node currLink = (Node)linkIt.next();

                if (!lg.isIn(currLink)) {
                    linkFlargest = false;
                }
            }
            if (linkFlargest) {
                newSibs.add(curr);
            }
            else if (!curr.linkNodes.isEmpty())
                os.add(curr);
        }
    }

    private static List<Node> subTrees(MatchedNodePair largest) {
        if (largest.node1.isRoot() && largest.node2.isRoot())
            throw new IllegalArgumentException("Cannot calculate subtrees with two roots.");

        List<Node> result = new ArrayList<>();
        List<Node> combined = combineSubTrees(largest);

        if(combined.isEmpty()) {
            return new ArrayList<>();
        } else {
            Node node = combined.get(0);
            String curr = node.getVenn();

            if ((!node.isLeaf()) && ((largest.node1.nSiblings() != 1) || (largest.node2.nSiblings() != 1))) {
                curr = curr.substring(1, curr.length() - 1);
            }

            String curr1 = "(" + largest.node1.getVenn() + curr + ")";
            String curr2 = "(" + largest.node2.getVenn() + curr + ")";
            Node newOne = new Node(null, buildUp(largest.node1.parent, curr1));
            Node newTwo = new Node(null, buildUp(largest.node2.parent, curr2));

            if ((!largest.node1.isRoot()) && (!largest.node2.isRoot())) {
                List<Node> newResult = combine(newOne, newTwo);
                result.addAll(newResult);
            }
            else if (!largest.node1.isRoot()) {
                result.add(newOne);
            }
            else if (!largest.node2.isRoot()) {
                result.add(newTwo);
            }
            return result;
        }
    }

    /**
     *
     * @param  h List
     * @return The largest MatchedNodePair in h, based on the length of n1's Venn diagram
     */
    private static List<MatchedNodePair> getLargestNodePairs(List<MatchedNodePair> h) {
        List<MatchedNodePair> result = new ArrayList<>();
        int size = -1;

        for(MatchedNodePair curr : h){
            if (curr.nodesSize() > size) {
                size = curr.nodesSize();
                result.clear();
                result.add(curr);
            } else if (curr.nodesSize() == size) {
                result.add(curr);
            }
        }
        return result;
    }

    private static Node combineAtRoot(Node node1, Node node2) { // CONVERTED
        List<Node> uniqueNodes = new ArrayList<>();
        for(Node curr : node1.children) {
            if (!uniqueNodes.contains(curr))
                uniqueNodes.add(curr);
        }
        for(Node curr : node2.children) {
            if (!uniqueNodes.contains(curr))
                uniqueNodes.add(curr);
        }
        if (uniqueNodes.size() < (node1.children.size() + node2.children.size())) {
            StringBuilder result_venn = new StringBuilder();
            result_venn.append("(");
            for (Node unique : uniqueNodes) {
                result_venn.append(unique);
            }
            result_venn.append(")");
            return new Node(null, result_venn.toString());
        } else {
            return new Node(null, "(" + node1.venn + node2.getVenn() + ")");
        }
    }

    private static List<MatchedNodePair> getLinkedNodePair(List<Node> h) {
        List<MatchedNodePair> result = new ArrayList<>();
        for (Node curr : h) {
            if (curr.hasLink())
                result.add(new MatchedNodePair(curr, curr.getLink()));
        }
        return result;
    }

    private static List<List<MatchedNodePair>> reviewPolytomies(List<MatchedNodePair> nodePairs) {
        List<List<MatchedNodePair>> result = new ArrayList<>();
        for (MatchedNodePair curr : nodePairs) {
            List<MatchedNodePair> currResult = new ArrayList<>();
            for (MatchedNodePair inCurr : nodePairs) {
                if (inCurr == curr) {
                    continue;
                }
                List<Node> linkedNodes = inCurr.node2.linkNodes;
                boolean found = false;
                for (Node currNP : linkedNodes) {
                    if (currNP == curr.node2) {
                        found = true;
                        inCurr.node2.linkNode = currNP;
                        break;
                    }
                }
                if (found) {
                    currResult.add(inCurr);
                }
            }
            if (!currResult.isEmpty()) {
                currResult.add(curr);
                result.add(currResult);
            }
        }
        return result;
    }

    private static List<MatchedNodePair> addPolys(List<MatchedNodePair> h) {
        List<MatchedNodePair> setResult = new ArrayList<>();
        MatchedNodePair first = null;
        StringBuilder result = new StringBuilder();
        for (MatchedNodePair nodes : h) {
            if (first == null) {
                first = nodes;
                result.append(first.node2.getVenn());
            }
            result.append(nodes.node1.getVenn());
        }
        for (MatchedNodePair nodes : h) {
            MatchedNodePair toAdd = new MatchedNodePair((nodes).node2, new Node(null, "(" + result + ")"));
            setResult.add(toAdd);
        }
        return setResult;
    }

    private static List<Node> getUniqueNodes(List<MatchedNodePair> nodePairs) {
        /* TODO: Figure out what this is actually supposed to do.
         * By the name it should return the set of nodes in the node pairs that
         * are unique, but if it actually does that, some tests fail. For now, just
         * adding all the nodes into the result seems to work.
         */
        List<Node> result_list = new ArrayList<>();
        for (MatchedNodePair curr : nodePairs) {
            result_list.add(curr.node1);
            result_list.add(curr.node2);
        }
        return result_list;
    }

    private static List<Node> combineSubTrees(MatchedNodePair np) {
        List<Node> commonSiblings = new ArrayList<>();

        List<Node> node1Siblings = new ArrayList<>(np.node1.getSiblings());
        Iterator it1 = node1Siblings.iterator();
        List<Node> node2Siblings = new ArrayList<>(np.node2.getSiblings());
        Iterator it2 = node2Siblings.iterator();

        StringBuilder siblingsVenn = new StringBuilder();

        if (!node1Siblings.isEmpty()) {
            siblingsVenn.append(((Node) it2.next()).getVenn());
        } else {
            if (it1.hasNext()) {
                siblingsVenn.append(((Node) it1.next()).getVenn());
                if (it1.hasNext()) {
                    String temp = siblingsVenn.toString();
                    siblingsVenn = new StringBuilder("(" + temp);
                    while (it1.hasNext()) {
                        siblingsVenn.append(((Node) it1.next()).getVenn());
                    }
                    siblingsVenn.append(")");
                }
                commonSiblings.add(new Node(null, siblingsVenn.toString()));
                return commonSiblings;
            }
            return commonSiblings;
        }

        if (it2.hasNext()) {
            String temp = siblingsVenn.toString();
            siblingsVenn = new StringBuilder("(" + temp);
            while (it2.hasNext()) {
                siblingsVenn.append(((Node) it2.next()).getVenn());
            }
            siblingsVenn.append(")");
        }

        Node nodesibs1 = new Node(null, siblingsVenn.toString());
        if (!it1.hasNext()) {
            commonSiblings.add(nodesibs1);
            return commonSiblings;
        }
        StringBuilder siblings2 = new StringBuilder();
        if (it1.hasNext()) {
            siblings2.append(((Node) it1.next()).getVenn());
            if (it1.hasNext()) {
                String temp = siblings2.toString();
                siblings2 = new StringBuilder("(" + temp);
                while (it1.hasNext()) {
                    siblings2.append(((Node) it1.next()).getVenn());
                }
                siblings2.append(")");
            }
        }

        Node nodeSibs2 = new Node(null, siblings2.toString());
        List<Node> result = Node.combine(nodesibs1, nodeSibs2);
        return result;
    }

    private static List<MatchedNodePair> combineNew(MatchedNodePair np) {
        List<MatchedNodePair> setResult = new ArrayList<>();
        Node resultNode1 = new Node(null, np.node2.getRoot().getVenn());
        String added = "(" + np.node2.getVenn() + np.node1.getVenn() + ")";

        for (Node linkNode : np.node2.linkNodes) {
            String result = buildUp(linkNode, added);
            Node resultNode2 = new Node(null, result);
            setResult.add(new MatchedNodePair(resultNode1, resultNode2));
        }
        return setResult;
    }

    private static String buildUp(Node n, String s) {
        if (n.parent == null) {
            return s;
        } else {
            StringBuilder result = new StringBuilder();
            result.append(s);
            for(Node sibling : n.getSiblings()){
                result.append(sibling.getVenn());
            }
            return buildUp(n.parent, "(" + result + ")");
        }
    }

    private static void log(String s){
        System.out.println(s);
    }
}
